/*
 * PjOde.java
 *
 *
 * Created on December 29, 2004, 6:42 PM
 * 1st-4th order RK option added 11/3/05
 */

package GrapherODE;

import java.awt.*;
import java.awt.Color;
import jv.function.PuFunction;
import jv.project.PvPickEvent;
import jv.object.PsDebug;
import jv.project.PjProject;
import jv.vecmath.PdVector;
import jv.vecmath.PiVector;
import jv.number.PdVector_IP;
import jv.project.PvDisplayIf;
import jv.project.PvViewerIf;
import jv.project.PvCameraIf;
import jv.number.PuDouble;
import jv.number.PuInteger;
import jv.vecmath.PdMatrix;
import jv.geom.PgPolygon;
import jv.geom.PgPolygonSet;
import jv.geom.PgElementSet;
import jv.thirdParty.ruler.*;
/**
 *
 * @author David
 */
public class PjOde extends PjProject {
    protected           Panel                       pDisplay;
    protected           int                         m_order = 1;
    protected           int                         m_numeqns = 1;
    protected           String[]                    m_indepVars = {"y","t"};
    protected           String[]                    m_depVars = {"dy"};
    protected           String[]                    m_eqns = {"-y"};
    protected           PuDouble                    m_tIC;
    protected           PuInteger                   m_tSteps;
    protected           PuDouble                    m_tStepSize;
    protected           PuDouble                    m_tMin;
    protected           PuDouble                    m_tMax;
    protected           PdVector                    m_ICs;
    protected           PdVector_IP                 m_pICs;
    protected           PuFunction                  m_func;
    protected           PvDisplayIf                 m_disp2D;
    protected           PvDisplayIf                 m_disp3D;
    protected           PgPolygon[]                 m_2dgraphs;
    protected           PgPolygonSet                m_2dpolys;
    protected           PgPolygon                   m_3dgraphs;
    protected           PgVecFld                    m_vecfld;
    protected           int                      m_2dx = 1;
    protected           int[]                    m_2dy = new int[] {0};
    protected           boolean                     m_do2d = false;
    protected           int                      m_3dx = 1;
    protected           int                      m_3dy = 0;
    protected           int                      m_3dz = 0;
    protected           boolean                     m_do3d = false;
    protected           boolean                 m_autoScale2D = false;
    protected           boolean                 m_dovf = false;
    protected           boolean                 m_bMultiple = false;
    protected           PuDomain                m_clipbounds;
    protected           PgElementSet		m_blackboard;
    protected           PgAxes                  axes;
    protected           int                     m_RKorder = 4;

    protected           double[][]                  m_points;  //calculated points from R-K
    protected           Color[]                     m_colorlist = {Color.blue, Color.red, Color.green, Color.orange, Color.yellow, Color.cyan, Color.magenta, Color.pink};
    
    /** Creates a new instance of PjOde */
    public PjOde() {
          super("ODE Grapher");
          
          if (getClass() == PjOde.class) 
              init();
    }
    
    public void init() {
        super.init();
        
        m_points = new double[0][0];
        
        m_tMin  = new PuDouble("Min t",this);
         m_tMin.setDefBounds(-30.,30.,0.5,1.0);
         m_tMin.setDefValue(-5);
         m_tMin.init();
         
         m_tMax  = new PuDouble("Max t",this);
         m_tMax.setDefBounds(-30.,30.,0.5,1.0);
         m_tMax.setDefValue(5);
         m_tMax.init();
         
         m_tStepSize = new PuDouble("step size", this);
         m_tStepSize.setDefBounds(0.01, 10, .01, .1);
         m_tStepSize.setDefValue(0.2);
         m_tStepSize.init();
        
         m_ICs = new PdVector(m_order);
         m_pICs = new PdVector_IP();
         m_pICs.setParent(this);
         m_pICs.setVector(m_ICs);
        
        setupFunc(m_order,m_numeqns);
        
         m_2dpolys = new PgPolygonSet();
        m_3dgraphs = new PgPolygon();
        m_vecfld = new PgVecFld();
         pDisplay = new Panel();
        
        //set up x-y grid and window
        m_blackboard = new PgElementSet(2); 
        m_blackboard.setName("X-Y Plane");
	m_blackboard.showElements(false);
	m_blackboard.setGlobalEdgeColor(new Color(161, 161, 161));
	m_blackboard.setGlobalElementColor(Color.white);
                
                m_clipbounds = new PuDomain();
                m_clipbounds.setVars(new String[] {"Horiz","Vert"});
                m_clipbounds.setDefBnds(new double[] {-1000.0,1000.0,1.,5.});
                m_clipbounds.setDefVals(new double[] {-5.,5.});
                m_clipbounds.setNumDiscr(0);
                m_clipbounds.init();
                m_clipbounds.setParent(this);
        
    }
    
    /** sets display to show 2d and/or 3d views */
    public void setDP (boolean s2d, boolean s3d) {
        pDisplay.removeAll();
        if (s2d && s3d) {
            pDisplay.setLayout(new GridLayout(2, 1));
        } else if (s2d || s3d) {
            pDisplay.setLayout(new GridLayout(1, 1));
        } else {
            
        }
        if (s2d) {
            pDisplay.add((Component)get2DDisplay());
  //          setDisplay(m_disp2D);
        }
        if (s3d) {
            pDisplay.add((Component)get3DDisplay());
        }
        
        pDisplay.validate();
       
    }
   
    public void start() {
        super.start();
        m_do2d = true;
        drawXYgrid();
        doRK();
        m_disp2D.addGeometry(m_vecfld);
        
        
    }
    
    public boolean update(Object event) {
		
         if (event == this) {
            doRK();
            return super.update(this);
        }
              if (event == m_tMin) {
                    if (m_tMin.getValue() > m_ICs.getLastEntry()) {
                        m_tMin.setValue(m_ICs.getLastEntry());
                    }
                    doRK();
                    return true;
                }
                if (event == m_tMax) {
                    if (m_tMax.getValue() < m_ICs.getLastEntry()) {
                        m_tMax.setValue(m_ICs.getLastEntry());
                    }
                    doRK();
                    return true;
                }
                if (event == m_tStepSize) {
                    doRK();
                    return true;
                }

                if (event == m_func) {
                    doRK();
                    return true;
                }
                if (event == m_pICs) {
                    doRK();
                    return true;
                }
                if (event == m_clipbounds) {
                    drawXYgrid();
                    m_disp2D.update(m_disp2D);
                    return true;
                }
                return super.update(event);
	}
    
    
    /** Draws 2D graphs 
    *   Assumes runge_katta has already been called
    *
    *  @param   h   variable to be on horiz - index into m_indepVars
    *  @param   k   array of vars to be on vert - index into m_indepVars
    */
    public void make2Dgraphs() {
        int h = m_2dx;
        int[] k = m_2dy;
        
        boolean firsttime = false;
        if (m_2dpolys.getNumVertices() == 0) {
            m_2dpolys.setGlobalPolygonColor(Color.black);
            m_2dpolys.setGlobalPolygonSize(2);
            m_2dpolys.showPolygonColors(true);
            m_2dpolys.assurePolygonColors();
            m_2dpolys.showVertices(false);
            firsttime = true;
        }
        
        int ind = 0;
        int polystart = 0;
        if (m_bMultiple) {
            int curNumVerts = m_2dpolys.getNumVertices();
            polystart = m_2dpolys.getNumPolygons();
            m_2dpolys.setNumVertices(curNumVerts + k.length*m_points.length);
            m_2dpolys.setNumPolygons(polystart + k.length);
            ind = curNumVerts;
        } else {
            m_2dpolys.setNumVertices(k.length * m_points.length);
            m_2dpolys.setNumPolygons(k.length);
        }
        PdVector[] vertex = m_2dpolys.getVertices();
 
        for (int i=0;i<k.length;i++) {

            m_2dpolys.setPolygonColor(polystart + i,m_colorlist[i]);
            PiVector poly = m_2dpolys.getPolygon(polystart + i);
            poly.setSize(m_points.length);
            for (int j=0;j<m_points.length;j++) {
                vertex[ind].m_data[0]  = m_points[j][h];
                vertex[ind].m_data[1] = m_points[j][k[i]];
                vertex[ind].m_data[2] = 0;
                poly.m_data[j] = ind;
                ind++;
            }

        }
       
        if (firsttime) {m_disp2D.addGeometry(m_2dpolys);}
        m_2dpolys.setVisible(true);
        
        m_2dpolys.update(m_2dpolys);
        if (m_autoScale2D) {drawXYgrid();}
        m_disp2D.update(m_disp2D);
        
    }
    
    /**Draws 3D graph
     *  Assumes runge_katta has already been called
     *
     * @param   x   variable to be on x axis - index into m_indepVars
     * @param   y   variable to be on y axis - index into m_indepVars
     * @param   z   variable to be on z axis - index into m_indepVars
     */
    public void make3Dgraph() {
        int x = m_3dx;
        int y = m_3dy;
        int z = m_3dz;
        m_3dgraphs.setNumVertices(m_points.length);
        m_3dgraphs.assureEdgeColors();
        m_3dgraphs.setGlobalEdgeSize(1);
        m_3dgraphs.setGlobalEdgeColor(Color.black);
        
        m_3dgraphs.showVertices(false);
        
        PdVector[] vecs = new PdVector[m_points.length];
        for (int j=0;j<m_points.length;j++) {
                vecs[j] = new PdVector(m_points[j][x], m_points[j][y],m_points[j][z]);
        }
        m_3dgraphs.setVertices(vecs);
        m_3dgraphs.makeEdgeColorsFromZ();
        
        m_3dgraphs.setVisible(true);
        m_3dgraphs.update(m_3dgraphs);
       m_disp3D.fit();
        m_disp3D.update(m_disp3D);
    }
    
    
    /** sets up function */
    public void setupFunc(int order, int numeq) {
        if (numeq > 1) {  // if it's a system
            switch (numeq) {
                case 2: 
                    m_depVars = new String[] {"dx","dy"};
                    m_indepVars = new String[] {"x","y","t"};
                    m_eqns = new String[] {"-y","x"};
                    break;
                case 3:
                    m_depVars = new String[] {"dx","dy","dz"};
                    m_indepVars = new String[] {"x","y","z","t"};
                    m_eqns = new String[] {"-z+x","x-y","z+y"};
                    break;
            }
            m_ICs.setSize(numeq+1);
        } else { //if it's not a system
            switch (order) {
                case 1:
                    m_depVars = new String[] {"dy"};
                    m_indepVars = new String[] {"y","t"};
                    m_eqns = new String[] {"-y"};
                    break;
                case 2:
                     m_depVars = new String[] {"d2y"};
                    m_indepVars = new String[] {"y","dy","t"};
                    m_eqns = new String[] {"dy - 2*y"};
                    break;
                case 3:
                    m_depVars = new String[] {"d3y"};
                    m_indepVars = new String[] {"y","dy","d2y","t"};
                    m_eqns = new String[] {"d2y + 2*dy - 3*y"};
                    break;
                case 4:
                    m_depVars = new String[] {"d4y"};
                    m_indepVars = new String[] {"y","dy","d2y","d3y","t"};
                    m_eqns = new String[] {"d3y + d2y - dy - y"};
                    break;
             }
            m_ICs.setSize(order+1);
        }
        m_pICs.setVector(m_ICs);
        
        m_func = new PuFunction(m_indepVars.length,m_depVars.length);
        m_func.setName("ODE");
        m_func.setVariables(m_indepVars);
        m_func.setExpressionNames(m_depVars);
        m_func.setExpressions(m_eqns);
        m_func.setParent(this);
        
        
    }
    
    public PvDisplayIf get2DDisplay() {
		if (m_disp2D != null)
			return m_disp2D;
		// Get viewer and ask for another display
		PvViewerIf viewer = getViewer();
		// Create top window and add xygrid
		m_disp2D = viewer.newDisplay("2D Display", false);
		m_disp2D.setBackgroundColor(Color.white);
                m_disp2D.selectCamera(PvCameraIf.CAMERA_ORTHO_XY);
                m_disp2D.setMajorMode(PvDisplayIf.MODE_INITIAL_PICK);
                
                m_disp2D.addGeometry(m_blackboard);
                setDisplay(m_disp2D);
		return m_disp2D;
	}
    public PvDisplayIf get3DDisplay() {
		if (m_disp3D != null)
			return m_disp3D;
		// Get viewer and ask for another display
		PvViewerIf viewer = getViewer();
		// Create 3d window and add 3dgraph geometry
		m_disp3D = viewer.newDisplay("3D Display", false);
		m_disp3D.setBackgroundColor(Color.white);
                m_disp3D.addGeometry(m_3dgraphs);
		return m_disp3D;
	}
    
    
    public void doRK () {
       //version based on initial t, step size, num steps:
      //  m_points = runge_katta(m_func, m_ICs.getEntries(), m_tIC.getValue(), m_tIC.getValue()+m_tSteps.getValue()*m_tStepSize.getValue(), m_tStepSize.getValue());
      if (m_tMin.getValue() < m_tMax.getValue()) {
          double[][] forward;
          double[][] backward;
          
          forward = runge_katta(m_func, m_ICs.getEntries(), m_tMax.getValue(), m_tStepSize.getValue());
          backward = runge_katta(m_func, m_ICs.getEntries(), m_tMin.getValue(), -1*m_tStepSize.getValue());
          m_points = new double[forward.length + backward.length + 1][];
          for (int i=0;i<backward.length;i++) {
              m_points[i] = backward[backward.length-i-1]; 
          }
          m_points[backward.length] = m_ICs.getEntries();
          for (int i=0;i<forward.length;i++) {
              m_points[backward.length+1+i] = forward[i]; 
          }
 
      } else {
          m_points = new double[1][5];
          m_points[0][0] = 0;
          m_points[0][1] = 0;
          m_points[0][2] = 0;
          m_points[0][3] = 0;
          m_points[0][4] = 0;
          
      }
       
        if (m_do2d) { 
            make2Dgraphs();
            m_disp2D.selectGeometry(m_blackboard);
        }
        if (m_do3d) {make3Dgraph();}
        if (m_dovf) {
                m_vecfld.drawVF(new double[] {m_clipbounds.getMin(0),m_clipbounds.getMin(1),m_clipbounds.getMax(0),m_clipbounds.getMax(1)},m_func, m_2dx, m_2dy[0], m_ICs); 
                m_vecfld.update(m_vecfld);
        } 
    }
    
    /** returns array of calculated points using fourth-order Runge-Katta method. 
     *  Treats higher order equations as systems for processing.
     *
     * @param   func    PuFunction
     * @param   ics     double[] initial conditions for m_indepVars
     * @param   b       ending t value
     * @param   h       step size (can be negative for b < t initial)
     */
public double[][] runge_katta(PuFunction func, double[] ics, double b, double h) {
	double[] y = removet(ics);
        double t = ics[ics.length-1];
        
	int subdivs = (int) Math.ceil(Math.abs((b-t)/h));
	double[][] pointlist = new double[subdivs][ics.length+1];

	
	double[] k1, k2, k3, k4, k;


	for (int i=0; i < subdivs; i++) {
		k1 = evalfuncs(func, t, y);	
               
		k2 = evalfuncs(func, t+h/2, addmulta(y,k1,h/2));  
                
                if (m_RKorder == 1) { k = k1;}
                else if (m_RKorder == 2) { k = diva(adda(k1,k2),2);}
                else if (m_RKorder == 3) {
                    // k3 = eval at t+h, y - k1 + 2k2
                    k3 = evalfuncs(func,t+h,addmulta(addmulta(y,k1,-h),k2,2*h));
                    k = diva(adda(addmulta(k1,k2,4),k3),6);
                } else {
		   k3 = evalfuncs(func, t+h/2, addmulta(y,k2,h/2));
                   k4 = evalfuncs(func, t+h, addmulta(y,k3,h));
                   //does k = (k1 + 2*k2 + 2*k3 + k4)/6;
                   k = diva(adda(addmulta(k1,k2,2), addmulta(k4,k3,2)),6);
                }
                
                
                y = addmulta(y,k,h); //does y += h*k;
		t += h;
                
		pointlist[i] = appendt(y, t);  
	}
	return pointlist;
}

/** Divides array by constant */
public double[] diva (double[] a, double c) {
	double[] out = new double[a.length];
	for (int i=0; i<a.length;i++) {
		out[i] = a[i]/c;
	}
	return out;
	
}

/**  Adds two double arrays componentwise */
public double[] adda (double[] a, double[] b) {
	if (a.length != b.length) {
		return null;
	} else {
		double[] out = new double[a.length];
		for (int i=0; i<a.length;i++) {
			out[i] = a[i] + b[i];
		}
		return out;
	}
}

/** Adds a+c*b, where a and b are arrays, c is a scalar */
public double[] addmulta (double[] a, double[] b, double c) {
	if (a.length != b.length) {
		return null;
	} else {
		double[] out = new double[a.length];
		for (int i=0; i<a.length;i++) {
			out[i] = a[i] + c*b[i];
		}
		return out;
	}
}

/** appends an entry to an array of doubles */
public double[] appendt (double[] a, double t) {
	double[] out = new double[a.length + 1];
	for (int j=0; j < a.length; j++) {
		out[j] = a[j];
	}
	out[a.length] = t;
        return out;
}

/** removes an entry from the end of an array of doubles */
public double[] removet (double[] a) {
	double[] out = new double[a.length - 1];
	for (int j=0; j < a.length-1; j++) {
		out[j] = a[j];
	}
	return out;
}

/** @param in	either values for y1, y2, y3, etc.. or
		values for y, dy, d2y, d3y, etc..
*/
public double[] evalfuncs (PuFunction func, double t, double[] in) {

	double[] out = new double[in.length];
	double[] toeval = appendt(in,t);
	
	if (func.getNumFunctions() == 1) {
		//for non-systems, convert to system y' = dy, dy' = d2y, etc..
		for (int j=0; j<in.length - 1; j++) {
			out[j] = in[j+1];
		}
		out[in.length-1] = func.eval(0,toeval);
 	} else {
 		// for systems
		func.eval(out,toeval);
	}
	return out;
}
    /** draws XY grid and axes */
    public void drawXYgrid() {
        double[] eps;
            eps = new double[2];
            double xmin = 1000;
            double xmax = -1000;
            double ymin = 1000;
            double ymax = -1000;
            
         if (m_autoScale2D) {   
            if ((m_points.length > 0) && (m_2dx > -1) && (m_2dy.length > 0)) {
                double x,y;
                for (int i=0; i< m_2dy.length;i++) {
                    for (int j=0; j < m_points.length; j++) {
                             y = m_points[j][m_2dy[i]];
                         if (y < ymin) { ymin = y;}
                         if (y > ymax) { ymax = y;}
                   }
                }
            
                for (int j=0; j < m_points.length; j++) {
                      x = m_points[j][m_2dx];
                    if (x < xmin) { xmin = x;}
                    if (x > xmax) { xmax = x;}
                }
                if (xmin == xmax) {
                    xmin = xmax -1;
                    xmax = xmin +2;
                }
                if (ymin == ymax) {
                    ymin = ymax -1;
                    ymax = ymin +2;
                }
            
            } else {
                xmin = -5;
                xmax = 5;
                ymin = -5;
                ymax = 5;
            }
        } else {
                xmin = m_clipbounds.getMin(0);
                xmax = m_clipbounds.getMax(0);
                ymin = m_clipbounds.getMin(1);
                ymax = m_clipbounds.getMax(1);
        }
            
            eps[0] = (xmax-xmin)/20.;
            eps[1] = (ymax-ymin)/20.;
            int[] hashings = new int[2];

            double[] majorhash = new double[2];
            for (int i =0; i<2; i++) {
                            if (eps[i] < .2) {
                                majorhash[i] = 2*eps[i];
                            } else if (eps[i] < 1) {
                                majorhash[i] = 1.0;
                            } else if (eps[i] < 2) {
                                majorhash[i] = 2.0;
                            } else {
                                majorhash[i] = 5*(Math.ceil(eps[i]/2.5));
                            }
            }
  
            double[] planebox = new double[4];
                       planebox[0] = majorhash[0]*(Math.floor(xmin/majorhash[0]));
                       planebox[1] = majorhash[1]*(Math.floor(ymin/majorhash[1]));
                       planebox[2] = majorhash[0]*(Math.ceil(xmax/majorhash[0]));
                       planebox[3] = majorhash[1]*(Math.ceil(ymax/majorhash[1]));
                       hashings[0] = (int) Math.floor((planebox[2]-planebox[0])/majorhash[0]) + 1;
                       hashings[1] = (int) Math.floor((planebox[3]-planebox[1])/majorhash[1]) + 1;
             
            PvDisplayIf disp = get2DDisplay();
            for (int i=0; i<2;i++) {
//                    disp.setClipBounds(i, new double[] {m_clipbounds.getMin(i)-eps[i], m_clipbounds.getMax(i)+eps[i]});
                      disp.setClipBounds(i, new double[] {planebox[i]-eps[i], planebox[i+2]+eps[i]});
            }
            axes = ((jv.viewer.PvDisplay)disp).getAxes();
			if (axes == null) {
                            axes = new PgAxes(2);
                            axes.setDisplay(disp);
                            axes.setMode(PgAxes.AXES_2DXY_CENTER);
                            
                            PdVector [] bndbox = PdVector.realloc(null, 2, 3);
//                            bndbox[0].set(xmin-eps[0],ymin-eps[1], 0.); 
//                            bndbox[1].set(xmax+eps[0],ymax+eps[1], 0.); 
                            bndbox[0].set(planebox[0]-eps[0],planebox[1]-eps[1],0.);
                            bndbox[1].set(planebox[2]+eps[0],planebox[3]+eps[1],0.);
                            axes.setBounds(bndbox);
                            axes.setEnabledAutoBounds(false);
                            axes.setEnabledAutoHashing(false);
                            axes.setNames(new String[] {m_indepVars[m_2dx], m_indepVars[m_2dy[0]], ""});
                            axes.showNames(true);
                            axes.setHashMode(PgRuler.HASH_FIXED); 
                            axes.setDisplay(disp);
                            ((jv.viewer.PvDisplay)disp).setAxes(axes);
                            disp.showAxes(true);
                            
                        } else {
                            axes.setEnabledAutoBounds(true);
                            PdVector [] bndbox = PdVector.realloc(null, 2, 3);
                            bndbox[0].set(planebox[0]-eps[0],planebox[1]-eps[1],0.);
                            bndbox[1].set(planebox[2]+eps[0],planebox[3]+eps[1],0.);
                              axes.setBounds(bndbox);
                            axes.setEnabledAutoBounds(false);
                        }    
            
                          
                      
                       axes.getAxes()[0].setFixedHashing(majorhash[0],majorhash[0]>2? majorhash[0]/5 : majorhash[0]/2);
                       axes.getAxes()[1].setFixedHashing(majorhash[1],majorhash[1]>2? majorhash[1]/5 : majorhash[1]/2);
                       axes.update(axes);

          
                       m_blackboard.computePlane(hashings[0],hashings[1], planebox[0] ,planebox[1],planebox[2] ,planebox[3] );
                       m_blackboard.update(m_blackboard);
                        
                disp.setEnabledBoxRatio(true);
         
                
                double xsize = xmax - xmin;
                double ysize = ymax - ymin;
                
                double scl = Math.min(xsize,ysize);
                
                PdVector br = new PdVector(scl/xsize,scl/ysize,1);
                 disp.setBoxRatio(br);
                
           
            double dist;
                PvCameraIf cam = disp.getCamera();
                
                PdVector interest = new PdVector(xmin+.5*xsize,ymin+.5*ysize,0);
                cam.setInterest(interest);
               
                dist = 1.2*scl;
                cam.setDist(dist);
                disp.update(cam);
                
        if (m_dovf) {
            m_vecfld.drawVF(new double[] {m_clipbounds.getMin(0),m_clipbounds.getMin(1),m_clipbounds.getMax(0),m_clipbounds.getMax(1)},m_func, m_2dx, m_2dy[0], m_ICs); 
            m_vecfld.update(m_vecfld);
        } 
    }
    
    public void pickInitial(PvPickEvent pos) {
		if (pos==null || pos.getGeometry()==null ||
			 pos.getGeometry()!=m_blackboard || m_2dy.length<1) {
//			if (PsDebug.WARNING) PsDebug.warning("select background grid and pick inside.");
                        
			return;
		}
		PdVector base = pos.getVertex();
                if (m_2dx == m_indepVars.length-1) {  //if we're picking a t-value
                    if ((base.m_data[0] < m_tMin.getValue()) || (base.m_data[0] > m_tMax.getValue())) {
                        return;
                    }
                }
                m_ICs.setEntry(m_2dx, base.m_data[0]);
                m_ICs.setEntry(m_2dy[0], base.m_data[1]);
		m_pICs.update(m_ICs);
		
		doRK();
	}
	public void dragInitial(PvPickEvent pos) { pickInitial(pos); }
}
